@@ -19,6 +19,7 @@
 module Control.Monad.Trans.Resource
     ( -- * Data types
       ResourceT
+    , ResIO
     , ReleaseKey
       -- * Unwrap
     , runResourceT
@@ -29,6 +30,13 @@ module Control.Monad.Trans.Resource
       -- * A specific Exception transformer
     , ExceptionT (..)
     , runExceptionT_
+    , runException
+    , runException_
+      -- * Registering/releasing
+    , allocate
+    , register
+    , release
+    , resourceMask
       -- * Type class/associated types
     , MonadResource (..)
     , MonadUnsafeIO (..)
@@ -70,6 +78,13 @@ import Control.Monad.Trans.State    ( StateT   )
 import Control.Monad.Trans.Writer   ( WriterT  )
 import Control.Monad.Trans.RWS      ( RWST     )
 
+import Control.Monad.Cont.Class   ( MonadCont (..) )
+import Control.Monad.Error.Class  ( MonadError (..) )
+import Control.Monad.RWS.Class    ( MonadRWS )
+import Control.Monad.Reader.Class ( MonadReader (..) )
+import Control.Monad.State.Class  ( MonadState (..) )
+import Control.Monad.Writer.Class ( MonadWriter (..) )
+
 import Data.Word (Word)
 
 import qualified Control.Monad.Trans.RWS.Strict    as Strict ( RWST   )
@@ -92,13 +107,13 @@ import qualified Control.Monad.ST.Lazy as LazyUnsafe
 
 import qualified Control.Monad.ST.Lazy as Lazy
 
-import Data.Functor.Identity (Identity)
+import Data.Functor.Identity (Identity, runIdentity)
 
 -- | A lookup key for a specific release action. This value is returned by
 -- 'register' and 'allocate', and is passed to 'release'.
 --
 -- Since 0.3.0
-newtype ReleaseKey = ReleaseKey Int
+data ReleaseKey = ReleaseKey !(I.IORef ReleaseMap) !Int
     deriving Typeable
 
 type RefCount = Word
@@ -119,7 +134,10 @@ data ReleaseMap =
 -- a release action will only ever be called once.
 --
 -- Since 0.3.0
-newtype ResourceT m a = ResourceT (I.IORef ReleaseMap -> m a)
+newtype ResourceT m a = ResourceT { unResourceT :: I.IORef ReleaseMap -> m a }
+
+-- | Convenient alias for @ResourceT IO@.
+type ResIO a = ResourceT IO a
 
 instance Typeable1 m => Typeable1 (ResourceT m) where
     typeOf1 = goType undefined
@@ -135,6 +153,32 @@ instance Typeable1 m => Typeable1 (ResourceT m) where
                 [ typeOf1 m
                 ]
 
+instance MonadCont m => MonadCont (ResourceT m) where
+  callCC f = ResourceT $ \i -> callCC $ \c -> unResourceT (f (ResourceT . const . c)) i
+
+instance MonadError e m => MonadError e (ResourceT m) where
+  throwError = lift . throwError
+  catchError r h = ResourceT $ \i -> unResourceT r i `catchError` \e -> unResourceT (h e) i
+
+instance MonadRWS r w s m => MonadRWS r w s (ResourceT m)
+
+instance MonadReader r m => MonadReader r (ResourceT m) where
+  ask = lift ask
+  local = mapResourceT . local
+
+mapResourceT :: (m a -> n b) -> ResourceT m a -> ResourceT n b
+mapResourceT f = ResourceT . (f .) . unResourceT
+
+instance MonadState s m => MonadState s (ResourceT m) where
+  get = lift get
+  put = lift . put
+
+instance MonadWriter w m => MonadWriter w (ResourceT m) where
+  tell   = lift . tell
+  listen = mapResourceT listen
+  pass   = mapResourceT pass
+
+
 -- | A @Monad@ which allows for safe resource allocation. In theory, any monad
 -- transformer stack included a @ResourceT@ can be an instance of
 -- @MonadResource@.
@@ -147,56 +191,69 @@ instance Typeable1 m => Typeable1 (ResourceT m) where
 --
 -- Since 0.3.0
 class (MonadThrow m, MonadUnsafeIO m, MonadIO m, Applicative m) => MonadResource m where
-    -- | Register some action that will be called precisely once, either when
-    -- 'runResourceT' is called, or when the 'ReleaseKey' is passed to 'release'.
+    -- | Lift a @ResourceT IO@ action into the current @Monad@.
     --
-    -- Since 0.3.0
-    register :: IO () -> m ReleaseKey
+    -- Since 0.4.0
+    liftResourceT :: ResourceT IO a -> m a
 
-    -- | Call a release action early, and deregister it from the list of cleanup
-    -- actions to be performed.
-    --
-    -- Since 0.3.0
-    release :: ReleaseKey -> m ()
+-- | Register some action that will be called precisely once, either when
+-- 'runResourceT' is called, or when the 'ReleaseKey' is passed to 'release'.
+--
+-- Since 0.3.0
+register :: MonadResource m => IO () -> m ReleaseKey
+register = liftResourceT . registerRIO
 
-    -- | Perform some allocation, and automatically register a cleanup action.
-    --
-    -- This is almost identical to calling the allocation and then
-    -- @register@ing the release action, but this properly handles masking of
-    -- asynchronous exceptions.
-    --
-    -- Since 0.3.0
-    allocate :: IO a -- ^ allocate
-             -> (a -> IO ()) -- ^ free resource
-             -> m (ReleaseKey, a)
+-- | Call a release action early, and deregister it from the list of cleanup
+-- actions to be performed.
+--
+-- Since 0.3.0
+release :: MonadIO m => ReleaseKey -> m ()
+release (ReleaseKey istate rk) = liftIO $ release' istate rk
 
-    -- | Perform asynchronous exception masking.
-    --
-    -- This is more general then @Control.Exception.mask@, yet more efficient
-    -- than @Control.Exception.Lifted.mask@.
-    --
-    -- Since 0.3.0
-    resourceMask :: ((forall a. ResourceT IO a -> ResourceT IO a) -> ResourceT IO b) -> m b
+-- | Perform some allocation, and automatically register a cleanup action.
+--
+-- This is almost identical to calling the allocation and then
+-- @register@ing the release action, but this properly handles masking of
+-- asynchronous exceptions.
+--
+-- Since 0.3.0
+allocate :: MonadResource m
+         => IO a -- ^ allocate
+         -> (a -> IO ()) -- ^ free resource
+         -> m (ReleaseKey, a)
+allocate a = liftResourceT . allocateRIO a
+
+-- | Perform asynchronous exception masking.
+--
+-- This is more general then @Control.Exception.mask@, yet more efficient
+-- than @Control.Exception.Lifted.mask@.
+--
+-- Since 0.3.0
+resourceMask :: MonadResource m => ((forall a. ResourceT IO a -> ResourceT IO a) -> ResourceT IO b) -> m b
+resourceMask = liftResourceT . resourceMaskRIO
 
 instance (MonadThrow m, MonadUnsafeIO m, MonadIO m, Applicative m) => MonadResource (ResourceT m) where
-    allocate acquire rel = ResourceT $ \istate -> liftIO $ E.mask $ \restore -> do
-        a <- restore acquire
-        key <- register' istate $ rel a
-        return (key, a)
+    liftResourceT = transResourceT liftIO
 
-    register rel = ResourceT $ \istate -> liftIO $ register' istate rel
+allocateRIO :: IO a -> (a -> IO ()) -> ResourceT IO (ReleaseKey, a)
+allocateRIO acquire rel = ResourceT $ \istate -> liftIO $ E.mask $ \restore -> do
+    a <- restore acquire
+    key <- register' istate $ rel a
+    return (key, a)
 
-    release rk = ResourceT $ \istate -> liftIO $ release' istate rk
+registerRIO :: IO () -> ResourceT IO ReleaseKey
+registerRIO rel = ResourceT $ \istate -> liftIO $ register' istate rel
 
-    resourceMask f = ResourceT $ \istate -> liftIO $ E.mask $ \restore ->
-        let ResourceT f' = f (go restore)
-         in f' istate
-      where
-        go :: (forall a. IO a -> IO a) -> (forall a. ResourceT IO a -> ResourceT IO a)
-        go r (ResourceT g) = ResourceT (\i -> r (g i))
+resourceMaskRIO :: ((forall a. ResourceT IO a -> ResourceT IO a) -> ResourceT IO b) -> ResourceT IO b
+resourceMaskRIO f = ResourceT $ \istate -> liftIO $ E.mask $ \restore ->
+    let ResourceT f' = f (go restore)
+     in f' istate
+  where
+    go :: (forall a. IO a -> IO a) -> (forall a. ResourceT IO a -> ResourceT IO a)
+    go r (ResourceT g) = ResourceT (\i -> r (g i))
 
-#define GO(T) instance (MonadResource m) => MonadResource (T m) where allocate a = lift . allocate a; register = lift . register; release = lift . release; resourceMask = lift . resourceMask
-#define GOX(X, T) instance (X, MonadResource m) => MonadResource (T m) where allocate a = lift . allocate a; register = lift . register; release = lift . release; resourceMask = lift . resourceMask
+#define GO(T) instance (MonadResource m) => MonadResource (T m) where liftResourceT = lift . liftResourceT
+#define GOX(X, T) instance (X, MonadResource m) => MonadResource (T m) where liftResourceT = lift . liftResourceT
 GO(IdentityT)
 GO(ListT)
 GO(MaybeT)
@@ -219,7 +276,7 @@ register' istate rel = I.atomicModifyIORef istate $ \rm ->
     case rm of
         ReleaseMap key rf m ->
             ( ReleaseMap (key - 1) rf (IntMap.insert key rel m)
-            , ReleaseKey key
+            , ReleaseKey istate key
             )
         ReleaseMapClosed -> throw $ InvalidAccess "register'"
 
@@ -240,9 +297,9 @@ instance Show InvalidAccess where
 instance Exception InvalidAccess
 
 release' :: I.IORef ReleaseMap
-         -> ReleaseKey
+         -> Int
          -> IO ()
-release' istate (ReleaseKey key) = E.mask $ \restore -> do
+release' istate key = E.mask $ \restore -> do
     maction <- I.atomicModifyIORef istate lookupAction
     maybe (return ()) restore maction
   where
@@ -253,7 +310,12 @@ release' istate (ReleaseKey key) = E.mask $ \restore -> do
                 ( ReleaseMap next rf $ IntMap.delete key m
                 , Just action
                 )
-    lookupAction ReleaseMapClosed = throw $ InvalidAccess "release'"
+    -- We tried to call release, but since the state is already closed, we
+    -- can assume that the release action was already called. Previously,
+    -- this threw an exception, though given that @release@ can be called
+    -- from outside the context of a @ResourceT@ starting with version
+    -- 0.4.4, it's no longer a library misuse or a library bug.
+    lookupAction ReleaseMapClosed = (ReleaseMapClosed, Nothing)
 
 stateAlloc :: I.IORef ReleaseMap -> IO ()
 stateAlloc istate = do
@@ -364,6 +426,18 @@ newtype ExceptionT m a = ExceptionT { runExceptionT :: m (Either SomeException a
 runExceptionT_ :: Monad m => ExceptionT m a -> m a
 runExceptionT_ = liftM (either E.throw id) . runExceptionT
 
+-- | Run an @ExceptionT Identity@ stack.
+--
+-- Since 0.4.2
+runException :: ExceptionT Identity a -> Either SomeException a
+runException = runIdentity . runExceptionT
+
+-- | Run an @ExceptionT Identity@ stack, but immediately 'E.throw' any exception returned.
+--
+-- Since 0.4.2
+runException_ :: ExceptionT Identity a -> a
+runException_ = runIdentity . runExceptionT_
+
 instance Monad m => Functor (ExceptionT m) where
     fmap f = ExceptionT . (liftM . fmap) f . runExceptionT
 instance Monad m => Applicative (ExceptionT m) where
@@ -397,6 +471,39 @@ instance MonadBaseControl b m => MonadBaseControl b (ExceptionT m) where
     liftBaseWith = defaultLiftBaseWith StE
     restoreM = defaultRestoreM unStE
 
+instance MonadCont m => MonadCont (ExceptionT m) where
+  callCC f = ExceptionT $
+    callCC $ \c ->
+    runExceptionT (f (\a -> ExceptionT $ c (Right a)))
+
+instance MonadError e m => MonadError e (ExceptionT m) where
+  throwError = lift . throwError
+  catchError r h = ExceptionT $ runExceptionT r `catchError` (runExceptionT . h)
+
+instance MonadRWS r w s m => MonadRWS r w s (ExceptionT m)
+
+instance MonadReader r m => MonadReader r (ExceptionT m) where
+  ask = lift ask
+  local = mapExceptionT . local
+
+mapExceptionT :: (m (Either SomeException a) -> n (Either SomeException b)) -> ExceptionT m a -> ExceptionT n b
+mapExceptionT f = ExceptionT . f . runExceptionT
+
+instance MonadState s m => MonadState s (ExceptionT m) where
+  get = lift get
+  put = lift . put
+
+instance MonadWriter w m => MonadWriter w (ExceptionT m) where
+  tell   = lift . tell
+  listen = mapExceptionT $ \ m -> do
+    (a, w) <- listen m
+    return $! fmap (\ r -> (r, w)) a
+  pass   = mapExceptionT $ \ m -> pass $ do
+    a <- m
+    return $! case a of
+        Left  l      -> (Left  l, id)
+        Right (r, f) -> (Right r, f)
+
 -- | A @Monad@ which can throw exceptions. Note that this does not work in a
 -- vanilla @ST@ or @Identity@ monad. Instead, you should use the 'ExceptionT'
 -- transformer in your stack if you are dealing with a non-@IO@ base monad.
@@ -408,6 +515,13 @@ class Monad m => MonadThrow m where
 instance MonadThrow IO where
     monadThrow = E.throwIO
 
+instance MonadThrow Maybe where
+    monadThrow _ = Nothing
+instance MonadThrow (Either SomeException) where
+    monadThrow = Left . E.toException
+instance MonadThrow [] where
+    monadThrow _ = []
+
 #define GO(T) instance (MonadThrow m) => MonadThrow (T m) where monadThrow = lift . monadThrow
 #define GOX(X, T) instance (X, MonadThrow m) => MonadThrow (T m) where monadThrow = lift . monadThrow
 GO(IdentityT)
